unit ElTools;

interface

uses Windows, SysUtils, classes, ElStrUtils;

function Max(a,b:integer):integer;
function Min(a,b:integer):integer;
function InRange(L,R,x:integer):boolean;
function RangesIntersect(L1,R1, L2,R2 : integer):boolean;
{$IFNDEF VER90}
function SubtractTimes(Time1, Time2 : TDateTime) : TDateTime;
{$ENDIF}
// the following two functions write and read the string in the following format:
// <string length><text>
function WriteStringToStream(S : TStream; Str : String) : boolean;
function ReadStringFromStream(S : TStream; var Str : String) : boolean;
// the following two functions write and read the string in the following format:
// <text>#13#10
procedure WriteTextToStream(S : TStream; Data : string);
function ReadTextFromStream(S : TStream; var Data : string) : boolean;

function DirExists(DirName : String) : boolean;

function CenterStr(Str:string;len:integer):string;
{ Centers string in a line }

function IsBIn(index:integer;storage:byte):boolean;

function encode_line(const buf; size:integer):string;

{$IFDEF VER90}
function CompareMem(Var1, Var2: pointer; size: integer): boolean; assembler;
{$ENDIF}

type TDirectMemoryStream = class (TMemoryStream)
       public
         procedure SetPointer(Ptr: Pointer; Size: Longint);
     end;

     TElMemoryStream = TDirectMemoryStream;

function GetsystemDir : string;

type TReducedDateTime = record
       Year,
       Month,
       Day,
       DOW,
       Hour,
       Min  : word;
     end;

function DateTimeToReduced(T:TDateTime):TReducedDateTime;
function ReducedToDateTime(T:TReducedDateTime):TDateTime;
function CompareReducedDT(T1,T2:TReducedDateTime):boolean;
// < -1
// = 0
// > +1
Function MakeReducedDT(Year,Month,Day,DOW,Hour,Min : word):TReducedDateTime;

function RunProgram(StartName, Params, StartDir : String):THandle;

procedure CenterRects(WS, WT, HS, HT : integer; var R : TRect);

function IncTime(ATime: TDateTime; Hours, Minutes, Seconds, MSecs: Integer): TDateTime;

function DaysPerMonth(AYear, AMonth: Integer): Integer;

procedure ElSystemTimeToTzSpecificLocalTime(lpTimeZoneInformation : PTimeZoneInformation;
                                            var lpUniversalTime  : TSystemTime;
                                            var lpLocalTime : TSystemTime);

function GetFormattedTimeString(ADate : TDateTime; Format : string) : string;
function DayNumber(AYear, AMonth, ADay : integer): integer;
function WeekNumber(AYear, AMonth, ADay : integer): integer;

function TimeInMask(CronMask : string; T:TReducedDateTime) : boolean;

var IsWin95, IsWinNT, IsWin2000, IsWin95OSR2, IsWin98 : boolean;

const ElementFormatList : array [1 .. 38] of string
                       = ('D', 'DD', 'w', '', 'ww', 'W', 'd', 'dd',
                          'm', 'mm', 'mmm', 'mmmm', 'y', 'yy',
                          'n', 'nn', 'Y', 'YY', 'x', 'xx', 'h',
                          'hh', 'H', 'HH', 'M', 'MM', 'S', 'SS',
                          '', '', '', 'tt', '|', '''', '~', '.', ':', 'TT');

implementation

function TimeInMask(CronMask : string; T:TReducedDateTime) : boolean;

  function InPart(S:String; Value:integer):boolean;
  var s1,s2:string;

  begin
    if s='*' then result:=true else
    begin
      s1:=s;
      while pos(',',s1)>0 do
      begin
        s2:=copy(s1,1,pos(',',s1)-1);
        delete(s1,1,pos(',',s1));
        result:= s2 = IntToStr(value);
        if result then exit;
      end;
      result:= s1 = IntToStr(value);
    end;
  end;

begin
  result := false;
  if not InPart(ExtractWord(CronMask,1),T.Min) then exit;
  if not InPart(ExtractWord(CronMask,2),T.Hour) then exit;
  if not InPart(ExtractWord(CronMask,3),T.Day) then exit;
  if not InPart(ExtractWord(CronMask,4),T.DOW) then exit;
  if not InPart(ExtractWord(CronMask,5),T.Month) then exit;
  result:=true;
end;

function GetSystemDir: string;
var
  Buffer: array[0..1023] of Char;
begin
  SetString(Result, Buffer, GetSystemDirectory(Buffer, SizeOf(Buffer)));
  if Result[Length(Result)] <> '\' then Result := Result + '\';
end;

function GetFormattedTimeString(ADate : TDateTime; Format : string) : string;

function DefDateSeparator : string;
    var S : string;
    begin
      SetLength(S, 255);
      GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_SDATE, PChar(S), 254);
      SetLength(S, StrLen(PChar(S)));
      result := S;
    end;

    function DefTimeSeparator : string;
    var S : string;
    begin
      SetLength(S, 255);
      GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_STIME, PChar(S), 254);
      SetLength(S, StrLen(PChar(S)));
      result := S;
    end;

    function LongDateToStr(var ST : TSystemTime) : string;
    var S : string;
    begin
      SetLength(S, 255);
      GetDateFormat(LOCALE_USER_DEFAULT, LOCALE_NOUSEROVERRIDE or DATE_LONGDATE, @ST, nil, PChar(S), 254);
      SetLength(S, StrLen(PChar(S)));
      result := S;
    end;

    function LongTimeToStr(var ST : TSystemTime) : string;
    var S : string;
        F : string;
        s1 : string;
        ff : integer;
    begin
      SetLength(S, 255);
      GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_ITIME, PChar(S), 254);
      SetLength(S, StrLen(PChar(S)));
      s1 := DefTimeSeparator;
      ff := 0;
      if StrToIntDef(S, 0) = 0
         then F := 'hh' + s1 + 'mm' + s1 + 'ss tt'
         else begin F := 'HH' + s1 + 'mm' + s1 + 'ss'; ff := TIME_FORCE24HOURFORMAT; end;

      SetLength(S, 255);
      GetTimeFormat(LOCALE_USER_DEFAULT, ff, @ST, PChar(F), PChar(S), 254);
      SetLength(S, StrLen(PChar(S)));
      result := S;
    end;

    function GetAMPMSign(Period : integer) : String;
    var S : string;
    begin
      SetLength(S, 255);
      if Period = 0
         then GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_S1159, PChar(S), 254)
         else GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_S2359, PChar(S), 254);
      SetLength(S, StrLen(PChar(S)));
      result := S;
    end;

var S1, S2, S3, S4 : string;
    i  : integer;
    Dt : TDateTime;
    ST : TSystemTime;
begin
  S1 := Format;
  S2 := '';
  i := 1;
  S4 := S1;
  DT := ADate;
  DateTimeToSystemTime(DT, ST);
  ST.wDayOfWeek := DayOfWeek(DT) - 1;
  while i <= Length(S1) do
  begin
    if S1[i] = '''' then
    begin
      if (i < Length(S1)) and (S1[i+1] = '''') then
      begin
        S2 := S2 + '''';
        inc(i);
      end else
      begin
        inc(i);
        while (i <= Length(S1)) do
        begin
          if S1[i] = '''' then break;
          S2 := S2 + S1[i];
          inc(i);
        end;
      end;
    end else
    if S1[i] = '~' then S2 := S2 + #9 else
    if S1[i] = ':' then S2 := S2 + DefTimeSeparator else
    if S1[i] = '.' then S2 := S2 + DefDateSeparator else
    if S1[i] = '|' then S2 := S2 + #13#10 else
    if Pos('nn', S4) = 1 then
    begin
      S3 := IntToStr(DayNumber(ST.wYear, ST.wMonth, ST.wDay));
      while Length(S3) < 3 do S3 := '0' + S3;
      S2 := S2 + S3;
      inc(i);
    end else
    if S1[i] = 'n' then S2 := S2 + IntToStr(DayNumber(ST.wYear, ST.wMonth, ST.wDay)) else

    if Pos('YY', S4) = 1 then
    begin
      S3 := IntToStr(WeekNumber(ST.wYear, ST.wMonth, ST.wDay));
      if Length(S3) = 1 then S3 := '0' + S3;
      S2 := S2 + S3;
      inc(i);
    end else
    if S1[i] = 'Y' then S2 := S2 + IntToStr(WeekNumber(ST.wYear, ST.wMonth, ST.wDay)) else
    if Pos('yy', S4) = 1 then
    begin
      S2 := S2 + IntToStr(ST.wYear);
      inc(i);
    end else
    if S1[i] = 'y' then S2 := S2 + IntToStr(ST.wYear mod 100) else
    if Pos('mmmm', S4) = 1 then
    begin
      S2 := S2 + LongMonthNames[ST.wMonth];
      inc(i, 3);
    end else
    if Pos('mmm', S4) = 1 then
    begin
      S2 := S2 + ShortMonthNames[ST.wMonth];
      inc(i, 2);
    end else
    if Pos('mm', S4) = 1 then
    begin
      S3 := IntToStr(ST.wMonth);
      if Length(S3) = 1 then S3 := '0' + S3;
      S2 := S2 + S3;
      inc(i);
    end else
    if S1[i] = 'm' then S2 := S2 + IntToStr(ST.wMonth) else
    if Pos('TT', S4) = 1 then
    begin
      S2 := S2 + Uppercase(GetAMPMSign(ST.wHour div 12));
      inc(i);
    end else
    if Pos('tt', S4) = 1 then
    begin
      S2 := S2 + Lowercase(GetAMPMSign(ST.wHour div 12));
      inc(i);
    end else
    if Pos('ww', S4) = 1 then
    begin
      S2 := S2 + LongDayNames[ST.wDayOfWeek + 1];
      inc(i);
    end else
    if S1[i] = 'w' then S2 := S2 + ShortDayNames[ST.wDayOfWeek + 1] else
    if S1[i] = 'W' then S2 := S2 + IntToStr(ST.wDayOfWeek) else
    if Pos('hh', S4) = 1 then
    begin
      if (ST.wHour = 0) or (ST.wHour = 12) then S3 := '12' else S3 := IntToStr(ST.wHour mod 12);
      if Length(S3) = 1 then S3 := '0' + S3;
      S2 := S2 + S3;
      inc(i);
    end else
    if S1[i] = 'h' then
    begin
      if (ST.wHour = 0) or (ST.wHour = 12) then S2 := S2 + '12' else S2 := S2 + IntToStr(ST.wHour mod 12);
    end else
    if Pos('HH', S4) = 1 then
    begin
      S3 := IntToStr(ST.wHour);
      if Length(S3) = 1 then S3 := '0' + S3;
      S2 := S2 + S3;
      inc(i);
    end else
    if S1[i] = 'H' then S2 := S2 + IntToStr(ST.wHour) else
    if Pos('MM', S4) = 1 then
    begin
      S3 := IntToStr(ST.wMinute);
      if Length(S3) = 1 then S3 := '0' + S3;
      S2 := S2 + S3;
      inc(i);
    end else
    if S1[i] = 'M' then S2 := S2 + IntToStr(ST.wMinute) else
    if Pos('SS', S4) = 1 then
    begin
      S3 := IntToStr(ST.wSecond);
      if Length(S3) = 1 then S3 := '0' + S3;
      S2 := S2 + S3;
      inc(i);
    end else
    if S1[i] = 'S' then S2 := S2 + IntToStr(ST.wSecond) else
    if Pos('dd', S4) = 1 then
    begin
      S3 := IntToStr(ST.wDay);
      if Length(S3) = 1 then S3 := '0' + S3;
      S2 := S2 + S3;
      inc(i);
    end else
    if S1[i] = 'd' then S2 := S2 + IntToStr(ST.wDay) else
    if Pos('xx', S4) = 1 then
    begin
      S2 := S2 + LongTimeToStr(ST);
      inc(i);
    end else
    if S1[i] = 'x' then S2 := S2 + TimeToStr(DT) else
    if Pos('DD', S4) = 1 then
    begin
      S2 := S2 + LongDateToStr(ST);
      inc(i);
    end else
    if S1[i] = 'D' then S2 := S2 + DateToStr(DT) else
    begin
      S2 := S2 + S1[i];
    end;
    inc(i);
    S4 := Copy(S1, i, Length(S1));
  end;
  Result := S2;
end;


function DayNumber(AYear, AMonth, ADay : integer): integer;
var i : integer;
begin
  Result := 0;
  for i := 1 to AMonth - 1 do Inc(Result, DaysPerMonth(AYear, i));
  Inc(Result, ADay);
end;

function WeekNumber(AYear, AMonth, ADay : integer): integer;
var
  FirstDate: TDateTime;
  FYearOffset, iWeekNum, FDay : integer;
begin
  FDay  := ADay;
  FYearOffset := DayNumber(AYear, AMonth, ADay);
  iWeekNum := ((FYearOffset  - 4) div 7) + 1;
  FirstDate := EncodeDate(AYear, 1, 1);
  if (FDay < 4) and (AMonth = 1) and (DayOfWeek(FirstDate) -1 > 4)
     then iWeekNum := WeekNumber(AYear - 1, 12, 31) else
     if iWeekNum <= 0 then iWeekNum := WeekNumber(AYear - 1, 12, 31);
  Result := iWeekNum;
end;

function IncTime(ATime: TDateTime; Hours, Minutes, Seconds, MSecs: Integer): TDateTime;
begin
  Result := ATime + (Hours div 24) + (((Hours mod 24) * 3600000 +
    Minutes * 60000 + Seconds * 1000 + MSecs) / MSecsPerDay);
end;

procedure CenterRects(WS, WT, HS, HT : integer; var R : TRect);
begin
  R.Left := max(WT div 2 - WS div 2, 0);
  R.Right := min(R.Left + WS, WT + R.Left);
  R.Top := max(HT div 2 - HS div 2, 0);
  R.Bottom := min(R.Top + HS, HT + R.Top);
end;

const bin2b64:string='ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';
      bin2uue:string='`!"#$%&''()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_';

function ReadTextFromStream;
var b : boolean;
    cs, cp, rr : integer;
    p : PChar;
begin
  cs := 256;
  cp := 0;
  Data := '';
  b := false;
  result := false;
  while not b do
  begin
    SetLength(Data, cp + cs);
    p := PChar(Data);
    inc(p, cp);
    rr := S.Read(p^, cs);
    if rr <> cs then
    begin
      b := true;
      SetLength(Data, cp + rr);
    end;
    if (Pos(#13#10, Data) > 0) then
    begin
      b := true;
      result := true;
      SetLength(Data, cp + rr);
      rr := rr - Pos(#13#10, StrPas(p)) - 1;
      S.Seek( -rr, soFromCurrent);
      SetLength(Data, Length(Data) -rr -2);
    end else inc(cp, cs);
  end;
end;

procedure WriteTextToStream;
var i : integer;
begin
  i := Length(Data);
  if (i < 2) or (not((Data[i-2] = #13) and (Data[i-1] = #10))) then Data := Data + #13#10;
  S.Write(PChar(Data)^, Length(Data));
end;

function encode_line(const buf; size:integer):string;
type ta_8u=packed array [0..65530] of byte;
var
  buff: ta_8u absolute buf;
  offset: shortint;
  pos1,pos2: byte;
  i: byte;
  out: string;

begin
  setlength(out, size*4 div 3 + 4);
  fillchar(out[1], size*4 div 3 +2,#0);   (* worst case *)
  offset:=2;
  pos1:=0;
  pos2:=2;
  out[pos2]:=#0;
  while pos1<size do
  begin
    if offset > 0 then
    begin
      out[pos2]:=char(ord(out[pos2]) or ((buff[pos1] and ($3f shl offset)) shr offset));
      offset:=offset-6;
      inc(pos2);
      out[pos2]:=#0;
    end
    else
    if offset < 0 then
    begin
      offset:=abs(offset);
      out[pos2]:=char(ord(out[pos2]) or ((buff[pos1] and ($3f shr offset)) shl offset));
      offset:=8-offset;
      inc(pos1);
    end
    else
    begin
      out[pos2]:=char(ord(out[pos2]) or ((buff[pos1] and $3f)));
      inc(pos2);
      inc(pos1);
      out[pos2]:=#0;
      offset:=2;
    end;
  end;
  if offset=2 then dec(pos2);
  for i:=2 to pos2 do
    out[i]:=bin2uue[ord(out[i])+1];
  encode_line:=copy(out,1,pos2);
end;

function DirExists(DirName : String) : boolean;
var SRec : TSearchRec;
begin
  result := false;
  if FindFirst(DirName, faAnyFile, SRec) = 0 then
  begin
    if (SRec.Attr and faDirectory) >0 then result := true;
  end;
  FindClose(SRec);
end;

function RunProgram;
var
    s:string;
    AppName:PChar;
    FParams :PChar;
    CrFlags:DWORD;
    CurDir :PChar;
    StartInfo:TStartupInfo;
    ProcInfo:TProcessInformation;
    res:boolean;

begin
  if IsWinNt then
  begin
    s:=StartName+#32+Params;
    AppName:=nil;
  end
  else
  begin
    s:=StartName;
    GetMem(AppName,260);
    StrPCopy(AppName,s);
    s:=Params;
  end;
  GetMem(FParams,260);
  StrPCopy(FParams,s);
  GetMem(CurDir,260);
  s:=StartDir;
  StrPCopy(CurDir,s);
  // Set process specific flags
  CrFlags:=NORMAL_PRIORITY_CLASS+CREATE_DEFAULT_ERROR_MODE;
  // set Startup information
  FillMemory(@StartInfo,SizeOf(TStartupInfo),0);
  with StartInfo do
  begin
    cb:=SizeOf(TStartupInfo);
    lpDesktop:=nil;
    lpTitle  :=nil;
    dwFlags  := STARTF_USESHOWWINDOW;
    wShowWindow := SW_SHOW;
  end; // with
  res:=CreateProcess(AppName, FParams, nil, nil, false, CrFlags, nil,
                  CurDir, StartInfo, ProcInfo);
  FreeMem(FParams,260);
  FreeMem(CurDir,260);
  if AppName<>nil then FreeMem(AppName,260);
  result:=0;
  if res then result:=ProcInfo.hProcess;
end;

procedure TDirectMemoryStream.SetPointer(Ptr: Pointer; Size: Longint);
begin
  inherited;
  Position := 0;
end;

Function MakeReducedDT(Year,Month,Day,DOW,Hour,Min : word):TReducedDateTime;
begin
  result.Year:=year;
  result.month:=month;
  Result.Day:=day;
  result.DOW:=dow;
  result.Hour:=hour;
  result.Min:=min;
end;

{$IFDEF VER90}

function CompareMem(Var1, Var2: pointer; size: integer): boolean; assembler;
asm
  push    esi
  push    edi
  xor     eax,eax
  mov     esi, var1
  mov     edi, var2
  mov     ecx, size
  shr     ecx,2
  mov     edx, size
  and     edx,3
  repe    cmpsd
  jne     @@2
  mov     ecx,edx
  repe    cmpsb
  jne     @@2
@@1:
  inc     eax
@@2:
  pop     edi
  pop     esi
end;

{$ENDIF}

function CompareReducedDT;
begin
  result:=CompareMem(@T1,@T2,SizeOf(TReducedDateTime));
end;

function DateTimeToReduced(T:TDateTime):TReducedDateTime;
var i,j:word;
begin
  DecodeDate(T,Result.Year,Result.Month,Result.Day);
  DecodeTime(T,Result.Hour,Result.Min,i,j);
  Result.DOW:=DayOfWeek(T);
end;

function ReducedToDateTime(T:TReducedDateTime):TDateTime;
begin
  try
    result:=EncodeDate(T.Year,T.Month,T.Day)+EncodeTime(T.Hour,T.Min,0,0);
  except
    on E:EConvertError do
    begin
      Result:=Now;
    end;
  end;
end;
                                 
procedure DefineOS;
var VerInfo:TOsVersionInfo;
begin
  VerInfo.dwOsVersionInfoSize:=SizeOf(VerInfo);
  GetVersionEx(VerInfo);
  IsWin95 := VerInfo.dwPlatformId = VER_PLATFORM_WIN32_WINDOWS;
  IsWin95OSR2 := (VerInfo.dwPlatformID = VER_PLATFORM_WIN32_WINDOWS) and (Loword(VerInfo.dwBuildNumber)>=1000);
  IsWinNt := VerInfo.dwPlatformId = VER_PLATFORM_WIN32_NT;
  IsWin2000 := (VerInfo.dwPlatformId = VER_PLATFORM_WIN32_NT) and (VerInfo.dwMajorVersion >= 5);
end;

function IsBIn;
begin
 result := ((Storage shr (index-1)) and 1)=1;
end;

Function CenterStr(Str:string; Len:integer):string;
var S:String;
    L:integer;
    ls : integer;
begin
 s:='';
 result:='';
 if len<Length(str) then Exit;
 l:=(Len-Length(Str)) div 2;
 while l > 0 do
 begin
  s:=s+' ';
  dec(l);
 end;
 ls := Length(Str);
 for l:=1 to ls do s:=s+str[L];
 CenterStr:=s;
end;

function InRange(L,R,x:integer):boolean;
begin
  result:=true;
  if (x<L) or (x>R) then result:=false;
end;

function Max(a,b:integer):integer;
begin
  if a<b then result:=b else result:=a;
end;

function Min(a,b:integer):integer;
begin
  if a>b then result:=b else result:=a;
end;
{$IFNDEF VER90}
function SubtractTimes(Time1, Time2 : TDateTime) : TDateTime;
var TT : TSystemTime;
    TFT1, TFT2, TFT3 : TFileTime;
    h1, h2, l1, l2 : integer;
begin
  begin
    DateTimeToSystemTime(Time1, TT);
    SystemTimeToFileTime(TT, TFT1);
    DateTimeToSystemTime(Time2, TT);
    SystemTimeToFileTime(TT, TFT2);
    h1 := TFT1.dwHighDateTime;
    h2 := TFT2.dwHighDateTime;
    l1 := TFT1.dwLowDateTime;
    l2 := TFT2.dwLowDateTime;
    if (h1 <  h2) or
       ((h1 = h2) and (l1 < l2)) then result := 0 else
       begin
         asm
           push eax
           mov eax, l1
           cmp eax, l2
           jl @1
           mov eax, l1
           sub eax, l2
           sub eax, $FFFFFFFF
           mov TFT3.dwLowDateTime, eax
           jmp @2
          @1:
          mov eax, l1
          sub eax, l2
          mov TFT3.dwLowDateTime, eax
          @2:
           // subtracting high-order dwords
           mov eax, h1
           sub eax, h2
           mov TFT3.dwHighDateTime, eax
           pop eax
         end;
         FileTimeToSystemTime(TFT3, TT);
         TT.wYear := 299 + TT.wYear ;
         result := SystemTimeToDateTime(TT);
       end;
    exit;
  end;
end;
{$ENDIF}
Function RangesIntersect(L1,R1, L2,R2 : integer):boolean;
begin
  if (Min(L1,R1)>Max(L2,R2)) or (Min(L2,R2)>Max(L1,R1))
     then result:=false
     else result:=true;
end;

function WriteStringToStream(S : TStream; Str : String) : boolean;
var i : integer;
begin
  i := Length(Str);
  try
    S.WriteBuffer(i, sizeof(integer));
    if i >0 then
      S.WriteBuffer(Str[1], i);
    result := true;
  except
    result := false;
  end;
end;

function ReadStringFromStream(S : TStream; var Str : String) : boolean;
var SS : String;
    i : integer;
begin
  i := Length(Str);
  try
    S.ReadBuffer(i, sizeof(integer));
    if i < 0 then raise Exception.Create('Invalid string header read from the stream');
    if i = 0 then Str := '' else
    begin
      SetLength(SS, i);
      S.ReadBuffer(SS[1], i);
      Str := SS;
    end;
    result := true;
  except
    result := false;
  end;
end;

function IsLeapYear(AYear: Integer): Boolean;
begin
  Result := (AYear mod 4 = 0) and ((AYear mod 100 <> 0) or (AYear mod 400 = 0));
end;

function DaysPerMonth(AYear, AMonth: Integer): Integer;
const
  DaysInMonth: array[1..12] of Integer = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
begin
  Result := DaysInMonth[AMonth];
  if (AMonth = 2) and IsLeapYear(AYear) then Inc(Result); { leap-year Feb is special }
end;

function IncDate(ADate: TDateTime; Days, Months, Years: Integer): TDateTime;
var
  D, M, Y: Word;
  Day, Month, Year: Longint;
begin
  DecodeDate(ADate, Y, M, D);
  Year := Y; Month := M; Day := D;
  Inc(Year, Years);
  Inc(Year, Months div 12);
  Inc(Month, Months mod 12);
  if Month < 1 then begin
    Inc(Month, 12);
    Dec(Year);
  end
  else if Month > 12 then begin
    Dec(Month, 12);
    Inc(Year);
  end;
  if Day > DaysPerMonth(Year, Month) then Day := DaysPerMonth(Year, Month);
  Result := EncodeDate(Year, Month, Day) + Days + Frac(ADate);
end;

{$ifdef VER90}

type PDayTable = ^TDayTable;
     TDayTable = array[1..12] of Word;

const
  MonthDays: array [Boolean] of TDayTable =
    ((31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31),
     (31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31));

function DoEncodeDate(Year, Month, Day: Word; var Date: TDateTime): Boolean;
var
  I: Integer;
  DayTable: PDayTable;
begin
  Result := False;
  DayTable := @MonthDays[IsLeapYear(Year)];
  if (Year >= 1) and (Year <= 9999) and (Month >= 1) and (Month <= 12) and
    (Day >= 1) and (Day <= DayTable^[Month]) then
  begin
    for I := 1 to Month - 1 do Inc(Day, DayTable^[I]);
    I := Year - 1;
    Date := I * 365 + I div 4 - I div 100 + I div 400 + Day - DateDelta;
    Result := True;
  end;
end;

function EncodeDate(Year, Month, Day: Word): TDateTime;
begin
  if not DoEncodeDate(Year, Month, Day, Result) then Raise exception.Create('Invalid argument for date encode.');
end;

procedure DivMod(Dividend: Integer; Divisor: Word;
  var Result, Remainder: Word);
asm
        PUSH    EBX
        MOV     EBX,EDX
        MOV     EDX,EAX
        SHR     EDX,16
        DIV     BX
        MOV     EBX,Remainder
        MOV     [ECX],AX
        MOV     [EBX],DX
        POP     EBX
end;

procedure DecodeDate(Date: TDateTime; var Year, Month, Day: Word);
const
  D1 = 365;
  D4 = D1 * 4 + 1;
  D100 = D4 * 25 - 1;
  D400 = D100 * 4 + 1;
var
  Y, M, D, I: Word;
  T: Integer;
  DayTable: PDayTable;
begin
  T := DateTimeToTimeStamp(Date).Date;
  if T <= 0 then
  begin
    Year := 0;
    Month := 0;
    Day := 0;
  end else
  begin
    Dec(T);
    Y := 1;
    while T >= D400 do
    begin
      Dec(T, D400);
      Inc(Y, 400);
    end;
    DivMod(T, D100, I, D);
    if I = 4 then
    begin
      Dec(I);
      Inc(D, D100);
    end;
    Inc(Y, I * 100);
    DivMod(D, D4, I, D);
    Inc(Y, I * 4);
    DivMod(D, D1, I, D);
    if I = 4 then
    begin
      Dec(I);
      Inc(D, D1);
    end;
    Inc(Y, I);
    DayTable := @MonthDays[IsLeapYear(Y)];
    M := 1;
    while True do
    begin
      I := DayTable^[M];
      if D < I then Break;
      Dec(D, I);
      Inc(M);
    end;
    Year := Y;
    Month := M;
    Day := D + 1;
  end;
end;

procedure DateTimeToSystemTime(DateTime: TDateTime; var SystemTime: TSystemTime);
begin
  with SystemTime do
  begin
    DecodeDate(DateTime, wYear, wMonth, wDay);
    DecodeTime(DateTime, wHour, wMinute, wSecond, wMilliseconds);
  end;
end;

function SystemTimeToDateTime(const SystemTime: TSystemTime): TDateTime;
begin
  with SystemTime do
    Result := EncodeDate(wYear, wMonth, wDay) +
      EncodeTime(wHour, wMinute, wSecond, wMilliSeconds);
end;

{$endif}

procedure ElSystemTimeToTzSpecificLocalTime(lpTimeZoneInformation : PTimeZoneInformation;
                                            var lpUniversalTime  : TSystemTime;
                                            var lpLocalTime : TSystemTime);
var UseStd : boolean;
    T1,
    T2,
    T3     : TSystemTime;
    DT1,
    DT2,
    DTF,
    FirstDate : TDateTime;
    FMonthOffset : integer;
    DayNum : integer;
    DIM : Integer;
begin
  UseStd := false;
  MoveMemory(@lpLocalTime, @lpUniversalTime, sizeof(TSystemTime));
  DTF := SystemTimeToDateTime(lpUniversalTime);
  DTF := IncTime(DTF, 0, -lpTimeZoneInformation.Bias, 0, 0);
  if lpTimeZoneInformation.StandardDate.wMonth <> 0 then
  begin
    MoveMemory(@T1,  @lpTimeZoneInformation.StandardDate, sizeof(TSystemTime));
    MoveMemory(@T2,  @lpTimeZoneInformation.DaylightDate, sizeof(TSystemTime));

    // converting stupid dates format to normal SystemTime
    T1.wYear := lpUniversalTime.wYear;
    T2.wYear := lpUniversalTime.wYear;

    FirstDate := EncodeDate(T1.wYear, T1.wMonth, 1);
    FMonthOffset := 2 - ((DateTimeToTimeStamp(FirstDate).Date mod 7 + 8) mod 7); { day of week for 1st of month }
    if FMonthOffset = 2 then FMonthOffset := -5;
    //if (FMonthOffset + T1.wDayOfWeek < 1) then inc(T1.wDay);
    DayNum := FMonthOffset + T1.wDayOfWeek + (T1.wDay) * 7;
    if (DayNum < 1) then inc(DayNum, 7);
    Dim := DaysPerMonth(T1.wYear, T1.wMonth);
    while (DayNum > Dim) do dec(DayNum, 7);
    T1.wDay := DayNum;

    FirstDate := EncodeDate(T2.wYear, T2.wMonth, 1);
    FMonthOffset := 2 - ((DayOfWeek(FirstDate) + 7) mod 7); { day of week for 1st of month }
    if FMonthOffset = 2 then FMonthOffset := -5;
    //if (FMonthOffset + T2.wDayOfWeek < 1) then inc(T2.wDay);
    DayNum := FMonthOffset + T2.wDayOfWeek + (T2.wDay) * 7;
    if (DayNum < 1) then inc(DayNum, 7);
    Dim := DaysPerMonth(T2.wYear, T2.wMonth);
    while (DayNum > Dim) do dec(DayNum, 7);
    T2.wDay := DayNum;
    repeat
      if (T1.wMonth <> T2.wMonth) then
      begin
        if (T1.wMonth < T2.wMonth) then UseStd := true;
        break;
      end;
      if (T1.wDay <> T2.wDay) then
      begin
        if (T1.wDay < T2.wDay) then UseStd := true;
        break;
      end;
      if (T1.wHour <> T2.wHour) then
      begin
        if (T1.wHour < T2.wHour) then UseStd := true;
        break;
      end;
      if (T1.wMinute <> T2.wMinute) then
      begin
        if (T1.wMinute < T2.wMinute) then UseStd := true;
        break;
      end;
      if (T1.wSecond <> T2.wSecond) then
      begin
        if (T1.wSecond < T2.wSecond) then UseStd := true;
        break;
      end;
    until true;
    if UseStd then
    begin
      MoveMemory(@T3, @T1, sizeof(TSystemTime));
      MoveMemory(@T1, @T2, sizeof(TSystemTime));
      MoveMemory(@T2, @T3, sizeof(TSystemTime));
    end;
    DT1 := SystemTimeToDateTime(T1);
    DT2 := SystemTimeToDateTime(T2);
    //SystemTimeToDateTime(lpUniversalTime);

    //if DTC > DT1 then DT2 := IncDate(DT2, 0, 0, 1) else
    if DTF < DT2 then DT1 := IncDate(DT1, 0, 0, -1);

    if (DTF < DT1) xor (UseStd)
       then DTF := IncTime(DTF, 0, -lpTimeZoneInformation.DaylightBias, 0, 0)
       else DTF := IncTime(DTF, 0, -lpTimeZoneInformation.StandardBias, 0, 0);
  end;
  DateTimeToSystemTime(DTF, lpLocalTime);
end;

initialization
begin
  DefineOS;
end;

end.
